def call_contract_eq(call, contract):
    return (
        call.function.target.label == contract or
        call.function.target.address == contract)


def call_method_eq(call, method):
    return call.function.name == method or call.function.selector == method


def call_depth_eq(call, depth):
    return call.prefix['depth'] == depth


def call_depth_lt(call, depth):
    return call.prefix['depth'] < depth



from functools import partial
from typing import cast, TYPE_CHECKING

from textual.command import Hit, Hits, Provider
from textual.message import Message

from rapidfuzz import process, fuzz, utils
from crucible.data.suite_tree import TestNode

if TYPE_CHECKING:
    from .suite import SuiteScreen


class TestSelected(Message):
    def __init__(self, node):
        self.node = node
        super().__init__()


class RunTestCommands(Provider):
    def parse_tests(self) -> list:
        screen: "SuiteScreen" = cast("SuiteScreen", self.screen)
        if screen.shadow_tree is None:
            return []
        return self.parse_test_node(screen.shadow_tree.root)

    def parse_test_node(self, node):
        if isinstance(node, TestNode):
            return [node]
        else:
            return [
                test_node
                for child in node.children
                for test_node in self.parse_test_node(child)
            ]

    async def startup(self) -> None:
        """Called once when the command palette is opened, prior to searching."""
        worker = self.app.run_worker(self.parse_tests, thread=True)
        self.test_nodes = await worker.wait()
        self.choices = [f"run {node.suite.contract} {
            node.test.test}" for node in self.test_nodes]

    async def search(self, query: str) -> Hits:
        """Search for Python files."""
        matcher = self.matcher(query)
        hits = process.extract(query, self.choices,
                               scorer=fuzz.token_set_ratio, limit=20,
                               processor=utils.default_process)

        for hit in hits:
            command, score, index = hit
            yield Hit(
                score,
                matcher.highlight(command),
                partial(self.screen.post_message, TestSelected(
                    self.test_nodes[index])),
            )



import pyparsing as pp

pp.ParserElement.set_default_whitespace_chars(' \t')

LPAR, RPAR, LBRACK, RBRACK, COLON, COMMA, NL = map(pp.Suppress, "()[]:,\n")

solc_version = pp.Combine(
    pp.Word(pp.nums) +
    pp.Literal(".") +
    pp.Word(pp.nums) +
    pp.Literal(".") +
    pp.Word(pp.nums)
)

compiling_started_line = pp.Group(
    pp.Literal("Compiling") +
    pp.common.number("count") +
    pp.Literal("files with Solc") +
    solc_version("version")
)("solc_start*")

compiling_finished_line = pp.Group(
    pp.Literal("Solc") +
    solc_version("version") +
    pp.Literal("finished in") +
    pp.Word(pp.alphanums + ".")("duration")
)("solc_finish*")

compiling_section = (
    (compiling_started_line + NL)[...] +
    (compiling_finished_line + NL)[...]
)

successful = pp.Group(pp.Literal("Compiler run successful!") + NL)("ok")
successuful_with_warnings = pp.Group(
    pp.Literal("Compiler run successful with warnings:") + NL
)("ok")
failed = (
    pp.Literal("Compiler run failed:") + NL
)("failed")
skipped = pp.Group(pp.Literal(
    "No files changed, compilation skipped") + NL)("skipped")

warning = pp.Group((
    pp.LineStart() +
    ((
        pp.Literal("Warning") +
        COLON
    ) |
        (
        pp.Literal("Warning") +
        LPAR +
        pp.common.number("code") +
        RPAR +
        COLON
    )) +
    pp.SkipTo(NL)("message") + NL
) + pp.SkipTo(
    pp.LineStart() + (
        pp.Literal("Warning") |
        pp.Literal("Error") |
        (pp.Literal("Ran") + pp.Word(pp.nums) + pp.Literal("tests for"))
    ) | pp.StringEnd()
)("hint"))("warnings*")

error = pp.Group((
    pp.LineStart() + pp.Literal("Error") +
    LPAR +
    pp.common.number("code") +
    RPAR +
    COLON +
    pp.SkipTo(NL)("message")
) + pp.SkipTo(
    pp.LineStart() + (
        pp.Literal("Warning") |
        pp.Literal("Error") |
        (pp.Literal("Ran") + pp.Word(pp.nums) + pp.Literal("test"))
    ) | pp.StringEnd()
)("hint"))("errors*")

compilation_success = pp.Group(compiling_section + (
    skipped | successful | (
        (successuful_with_warnings) +
        (warning | error)[...]
    )
))("compilation")

compilation_failed = pp.Group(compiling_section + (
    failed + (error | warning)[...]
))("compilation")

compilation_results = compilation_failed | compilation_success



import pyparsing as pp
from .traces import trace_line

pp.ParserElement.set_default_whitespace_chars(' \t')
LPAR, RPAR, LBRACK, RBRACK, COLON, COMMA, NL, PERIOD, SEMI = map(
    pp.Suppress, "()[]:,\n.;")


def plur(val):
    return pp.Literal(val) + pp.Optional(pp.Literal("s"))


test_status = (
    pp.Literal("PASS")("result") |
    pp.Literal("SKIP")("result") |
    (
        pp.Literal("FAIL")("result") +
        (pp.Literal(". Reason: ") | pp.Literal(": ")) +
        pp.SkipTo(RBRACK)("reason")
    )
)


test_result = (
    LBRACK + test_status + RBRACK +
    pp.Word(pp.alphanums + "_")("test") + LPAR + RPAR +
    LPAR + pp.Literal("gas:") + pp.common.number("gas") + RPAR
)

fuzz_test_result = (
    LBRACK + test_status + RBRACK +
    pp.Word(pp.alphanums + "_")("test") + LPAR + pp.SkipTo(RPAR + LPAR + pp.Literal("runs:"))("args") + RPAR +
    LPAR + pp.Literal("runs:") + pp.common.number("runs") + COMMA +
    pp.Literal("μ:") + pp.common.number("miu") + COMMA +
    pp.Literal("~:") + pp.common.number("average") + RPAR
)

invariant_test_result = (
    LBRACK + test_status + RBRACK +
    pp.Word(pp.alphanums + "_")("test") + LPAR + pp.SkipTo(RPAR + LPAR + pp.Literal("runs:"))("args") + RPAR +
    LPAR + pp.Literal("runs:") + pp.common.number("runs") + COMMA +
    pp.Literal("calls:") + pp.common.number("calls") + COMMA +
    pp.Literal("reverts:") + pp.common.number("reverts") + RPAR
)

logs = (
    pp.Literal("Logs") + COLON + NL +
    pp.SkipTo(pp.LineStart() + pp.Literal("Traces:"))("logs")
)

trace_group = pp.OneOrMore(pp.Group(trace_line + NL)("traces*"))

traces = (
    pp.Literal("Traces") + COLON + NL +
    pp.OneOrMore(pp.Group(trace_group + pp.Optional(NL))("trace_groups*"))
)

test = (
    (test_result | fuzz_test_result | invariant_test_result) + NL +
    pp.Optional(logs) + pp.Optional(traces)
)

suite_start = (
    pp.Literal("Ran") +
    pp.common.number("count") +
    plur("test") + pp.Literal("for") +
    pp.Word(pp.alphanums + "./:_-")("contract")
)

suite_end = (
    pp.Literal("Suite result:") + pp.Word(pp.alphas)("result") + PERIOD +
    pp.common.number("passed") + pp.Literal("passed") + SEMI +
    pp.common.number("failed") + pp.Literal("failed") + SEMI +
    pp.common.number("skipped") + pp.Literal("skipped") + SEMI +
    pp.Literal("finished in") + pp.Word(pp.alphanums + ".µ")("duration") +
    LPAR + pp.Word(pp.alphanums + ".µ")("cpu_time") +
    pp.Literal("CPU time") + RPAR
)

suite = (
    suite_start + NL +
    pp.OneOrMore(pp.Group(test)("tests*")) +
    suite_end + (NL | pp.StringEnd())
)

suites_end_line = (
    pp.Literal("Ran") + pp.common.number("count") +
    plur("test suite") + pp.Literal("in") +
    pp.Word(pp.alphanums + ".µ")("duration") +
    LPAR + pp.Word(pp.alphanums + ".µ")("cpu_time") +
    pp.Literal("CPU time") + RPAR + COLON +
    pp.common.number("passed") + plur("test") + pp.Literal("passed") + COMMA +
    pp.common.number("failed") + pp.Literal("failed") + COMMA +
    pp.common.number("skipped") + pp.Literal("skipped") +
    LPAR + pp.common.number("total") +
    pp.Literal("total") + plur("test") + RPAR + (NL | pp.StringEnd())
)

failing_test_suite = (
    pp.Literal("Encountered") + pp.common.number("count") +
    pp.Literal("failing") + plur("test") + pp.Literal("in") +
    pp.Word(pp.alphanums + "./:_-")("contract") + NL +
    pp.OneOrMore(pp.Group(test)("tests*"))
)

failing_tests_summary = pp.Group(
    pp.Literal("Failing tests:") + NL +
    pp.OneOrMore(pp.Group(failing_test_suite + NL)("suits*")) +
    pp.Literal("Encountered a total of") + pp.common.number("failed") +
    pp.Literal("failing") + plur("test") + pp.Literal(",") +
    pp.common.number("success") + plur("test") + pp.Literal("succeeded") + NL
)("failing")

suites = (
    pp.OneOrMore(pp.Group(suite)("suites*") + NL) +
    suites_end_line +
    pp.Optional(NL + failing_tests_summary)
)



from textual import work
from textual.containers import Container
from textual.screen import ModalScreen
from textual.reactive import reactive
from textual.widgets import LoadingIndicator, Static
from textual.worker import Worker, WorkerState

from crucible.runner import runner, RunConfig


class RunForgeTest(ModalScreen):
    tests: str
    run_config: RunConfig
    output = reactive(None)

    def __init__(
            self,
            run_config: RunConfig,
            title: str = "tests...",
            *args, **kwargs):
        super().__init__(*args, id="run-forge-test", **kwargs)
        self.run_config = run_config
        self.title = title

    def compose(self):
        yield Container(
            Static(f"Running {self.title}", classes="loading-text"),
            LoadingIndicator(classes="loading-indicator"),
            classes="modal-box"
        )

    def on_mount(self):
        self.load()

    @work(thread=True)
    async def load(self):
        return runner.run(self.run_config)

    def on_worker_state_changed(self, evt: Worker.StateChanged) -> None:
        if evt.state == WorkerState.SUCCESS:
            self.dismiss(evt.worker.result)



from os import system

from textual import on
from textual.app import App, ComposeResult
from textual.widgets import Button


class SuspendingApp(App[None]):

    def compose(self) -> ComposeResult:
        yield Button("Open the editor", id="edit")

    @on(Button.Pressed, "#edit")
    def run_external_editor(self) -> None:
        with self.suspend():
            system("vim")


if __name__ == "__main__":
    SuspendingApp().run()



from .shadow_tree import ShadowNode
from .trace_tree import CallNode
from .call_filters import call_contract_eq, call_method_eq


def isCall(contract, method=None):
    def filter(node: ShadowNode):
        if not isinstance(node, CallNode):
            return False
        return (
            call_contract_eq(node.call, contract) and
            method is None or call_method_eq(node.call, method))
    return filter



import pyparsing as pp
from .values import address, label, value, number

LPAR, RPAR, LBRACK, RBRACK, COLON, COMMA, NL = map(pp.Suppress, "()[]:,\n")

event_name = pp.Word(pp.alphas.upper(), pp.alphanums + "_")
event = (
    pp.CaselessKeyword("emit") +
    event_name("name") +
    LPAR +
    pp.dict_of(
        label("key") + COLON,
        pp.Group(value)("value") + pp.Optional(COMMA)
    )("args") +
    RPAR
)


def parse_prefix(tokens):
    depth = len(tokens.depth)
    if tokens.child != '' or tokens.last_child != '':
        depth += 1

    return {
        "depth": depth,
        "is_last_child": tokens.last_child != ''
    }


prefix = pp.Optional(
    pp.Literal("│")("depth*")[...] + pp.Optional(
        pp.Literal("├─")("child") |
        pp.Literal("└─")("last_child")
    ))("prefix").set_parse_action(parse_prefix)

target = pp.Group(address | label("label"))
gas = LBRACK + number("gas") + RBRACK
raw_args = LPAR + pp.Word(pp.hexnums) + RPAR
args = LPAR + (pp.Group(value)("args*") +
               pp.Optional(COMMA))[...] + RPAR

generic_function = pp.Group(
    target("target") +
    pp.Suppress("::") +
    (label("name") | pp.Word(pp.hexnums, exact=8)("selector")) + (
        raw_args("raw_args") |
        args
    ) + pp.Optional(LBRACK + label("call_type") + RBRACK))

expect_revert = pp.Group(
    pp.Group(pp.Literal("VM")("label"))("target") +
    pp.Suppress("::") +
    pp.Literal("expectRevert")("name") +
    LPAR + (
        pp.Group(label("name") + args)("custom_error") |
        pp.SkipTo(RPAR + pp.LineEnd())("reason")
    ) + RPAR
)


trace_parser = prefix + gas + (
    generic_function("function") |
    expect_revert("function"))

event_parser = pp.Optional(prefix("prefix")) + event

raw_event_parser = (
    prefix("prefix") + pp.Literal("emit topic 0:") + pp.Group(value)("topic0") +
    pp.Optional(NL + prefix + pp.Literal("topic 1:") + pp.Group(value)("topic1")) +
    pp.Optional(NL + prefix + pp.Literal("topic 2:") + pp.Group(value)("topic2")) +
    pp.Optional(NL + prefix + pp.Literal("topic 3:") + pp.Group(value)("topic3")) +
    pp.Optional(NL + prefix + pp.Literal("data:") + pp.Group(value)("data"))
)


return_value = (
    pp.Group(
        pp.Word(pp.nums)("count") +
        pp.Suppress("bytes of code")
    )("returned_bytes") |
    (
        pp.Group(value)("returned_values*") +
        pp.Optional(COMMA)
    )[...]
)

return_statment = pp.Group(
    pp.Suppress("← [Return]") +
    pp.Optional(return_value)
)("return")

revert_statement = pp.Group(
    pp.Suppress("← [Revert]") +
    pp.Optional(pp.SkipTo(pp.LineEnd())("reason"))
)("revert")

stop_statement = pp.Group(
    pp.Suppress("← [Stop]")
)("stop")

contract_create_parser = (
    prefix +
    gas +
    pp.Suppress("→ new") +
    pp.SkipTo("@")("contract_name") +
    pp.Suppress("@") +
    address("address"))

call_end_parser = prefix + \
    (stop_statement | revert_statement | return_statment)

trace_line = (
    trace_parser("call") |
    contract_create_parser("create") |
    event_parser("event") |
    raw_event_parser("raw_event") |
    call_end_parser("end")
)



from textual import log
from textual.widgets import Tree as TreeWidget

from crucible.data.shadow_tree import ShadowNode, ShadowTree


class Tree(TreeWidget):
    shadow_tree: ShadowTree
    nodes_by_shadow_node_id: dict

    BINDINGS = [
        ("l", "move_expand", "expand"),
        ("h", "move_collpase", "collpase"),
        ("k", "cursor_up", "up"),
        ("j", "cursor_down", "down"),
        # ("c", "copy", "copy"),
        # ("i", "inspect", "inspect"),
    ]

    def __init__(self, shadow_tree: ShadowTree, **kwargs):
        super().__init__(shadow_tree.root.label, shadow_tree.root.id, **kwargs)
        self.shadow_tree = shadow_tree
        self.nodes_by_shadow_node_id = {}

        self.root.expand()
        self.add_node(self.root, shadow_tree.root)
        if self.shadow_tree.focused is not None:
            log("Focused on init: ", self.shadow_tree.focused)
            self.call_after_refresh(
                self.move_to,
                self.nodes_by_shadow_node_id[self.shadow_tree.focused])

    def add_node(self, parent, node: ShadowNode):
        self.nodes_by_shadow_node_id[node.id] = parent
        for child in node.children:
            if (child.hidden):
                continue
            if child.children:
                child_node = parent.add(
                    child.label, child.id, expand=child.expanded)
            else:
                child_node = parent.add_leaf(child.label, child.id)
            self.add_node(child_node, child)

    def action_move_down(self):
        assert self.cursor_node is not None
        if self.cursor_node.is_last or not self.cursor_node.parent:
            return
        node_index = self.cursor_node.parent.children.index(self.cursor_node)
        self.move_to(self.cursor_node.parent.children[node_index + 1])

    def action_move_up(self):
        assert self.cursor_node is not None
        if self.cursor_node.parent is None:
            return
        node_index = self.cursor_node.parent.children.index(self.cursor_node)
        if node_index:
            self.move_to(self.cursor_node.parent.children[node_index - 1])

    def action_move_expand(self):
        assert self.cursor_node is not None
        if self.cursor_node.children:
            if self.cursor_node.is_expanded:
                self.move_to(self.cursor_node.children[0])
            else:
                self.cursor_node.expand()
                # self.call_after_refresh(
                #     self.move_to,
                #     self.cursor_node.children[0])

    def action_move_collpase(self):
        assert self.cursor_node is not None
        if self.cursor_node.is_expanded:
            self.cursor_node.collapse()

    def action_inspect(self):
        if self.cursor_node:
            node = self.shadow_tree.get_node(self.cursor_node.data)
            log(node.test)
            log(node.suite)

        assert self.cursor_node is not None
        if self.cursor_node.is_expanded:
            self.cursor_node.collapse()

    def move_to(self, node):
        self.scroll_to_node(node)
        self.move_cursor(node)

    def on_tree_node_expanded(self, evt):
        node = self.shadow_tree.get_node(evt.node.data)
        if node:
            node.toggle_expanded(True)

    def on_tree_node_collapsed(self, evt):
        node = self.shadow_tree.get_node(evt.node.data)
        if node:
            node.toggle_expanded(False)

    def on_tree_node_highlighted(self, evt):
        log("Focus: ", evt.node.data)
        self.shadow_tree.set_focused(evt.node.data)



import pyparsing as pp

from .compilation import compilation_success, compilation_failed
from .suites import suites

LPAR, RPAR, LBRACK, RBRACK, COLON, COMMA, NL, PERIOD, SEMI = map(
    pp.Suppress, "()[]:,\n.;")

output = compilation_failed | (
    compilation_success + pp.Optional(NL) + suites
)



from typing import Any, Optional

from textual import work
from textual.screen import Screen
from textual.widgets import Header, Footer, Static
from textual.reactive import reactive

from crucible.data.trace_tree import TraceTree
from crucible.data.suite_tree import TestNode, SuiteNode
from crucible.data.shadow_tree import ShadowTree, FilterKey
from crucible.data.node_filters import isCall
from crucible.runner import RunConfig
from crucible.screens.run_forge_test import RunForgeTest

from .tree import Tree
from .compilation_results import CompilationResults


class TestScreen(Screen):
    BINDINGS = [
        ("b", "back", "Back"),
        ("r", "run_test", "Rerun test"),
        ("f", "filters", "Filters")
    ]
    shadow_tree: Optional[ShadowTree] = None
    filters = reactive([], recompose=True)
    logs = None
    output: Any = reactive(None, recompose=True)

    def __init__(self, node, **kwargs):
        super().__init__(classes="test-screen", **kwargs)
        self.node = node
        if isinstance(node, TestNode):
            self.title = f"{node.suite.contract.split(
                ":")[-1]}::{node.test.test}"
        elif isinstance(node, SuiteNode):
            self.title = f"{node.suite.contract.split(":")[-1]}"

    def on_mount(self) -> None:
        self.action_run_test()

    def compose(self):
        yield Header(show_clock=True)
        if self.output and self.output.compilation.failed:
            yield CompilationResults(self.output, id="compilation")
        elif self.shadow_tree:
            yield Tree(self.shadow_tree, id="tree")
            if self.logs and self.logs.strip():
                yield Static("Logs:\n" + self.logs, id="logs")
        yield Footer()

    async def action_filters(self) -> None:
        if not self.shadow_tree:
            return
        if len(self.shadow_tree.filters) == 0:
            self.shadow_tree.add_filters({
                FilterKey('vm::label'): isCall("VM", "label"),
                FilterKey('vm::addr'): isCall("VM", "addr"),
                FilterKey('console::log'): isCall("console", "log"),
            })
        else:
            self.shadow_tree.filters = {}
        await self.recompose()
        self.call_after_refresh(self.focus_child, "tree")

    def action_back(self) -> None:
        self.app.pop_screen()

    @work
    async def action_run_test(self) -> None:
        run_config = self.get_run_config()
        output = await self.app.push_screen_wait(
            RunForgeTest(run_config, title=self.title or "")
        )

        self.output = output
        if self.output.compilation.failed:
            self.call_after_refresh(self.focus_child, "compilation")
        else:
            self.shadow_tree = TraceTree(output)
            self.logs = output.suites[0].tests[0].logs
            self.call_after_refresh(self.focus_child, "tree")

    def get_run_config(self) -> RunConfig:
        if isinstance(self.node, TestNode):
            return RunConfig("default",
                             self.node.suite.contract.split(':')[-1],
                             self.node.test.test,
                             None,
                             True)
        elif isinstance(self.node, SuiteNode):
            return RunConfig("default",
                             self.node.suite.contract.split(':')[-1],
                             None,
                             None,
                             True)
        else:
            raise ValueError(f"Unexpected node type: {type(self.node)}")

    def focus_child(self, id) -> None:
        self.set_focus(self.get_child_by_id(id))



from rich.text import Text
from textual import log
from textual.containers import ScrollableContainer
from textual.reactive import reactive
from textual.widgets import Static


class CompilationResults(ScrollableContainer):
    errors = reactive([], recompose=True)
    warnings = reactive([], recompose=True)

    BINDINGS = [
        ("j", "scroll_down", "Scroll down"),
        ("k", "scroll_up", "Scroll up")
    ]

    def __init__(self, output, **kwargs):
        super().__init__(**kwargs)
        self.errors = output.compilation.errors
        self.warnings = output.compilation.warnings
        log(self.errors)
        log(self.warnings)

    def compose(self):
        log(self.errors)
        log(self.warnings)
        if self.errors:
            for error in self.errors:
                yield Static(Text(f"Error [{error.code}]: {error.message}",
                                  style="red"))
                yield Static(error.hint.strip() + "\n")
        if self.warnings:
            for warning in self.warnings:
                code = warning.code
                if code == '':
                    code = hash(f"{warning.message}") % 1000000
                yield Static(Text(f"Warning [{code}]: {warning.message}",
                                  style="yellow"))
                yield Static(warning.hint.strip() + "\n")



from rich.text import Text


def light(text):
    return Text(text, style="dim light_slate_grey")


def bold(text):
    return Text(text, style="bold")


def default(text):
    return Text(text, style="default")


def keyword(text):
    return Text(text, style="bold yellow")


def event(text):
    return Text(text, style="bold blue")


def ret(text):
    return Text(text, style="bold dark_cyan")


def revert(text):
    return Text(text, style="bold red")



import pyparsing as pp
ppu = pp.unicode

LPAR, RPAR, LBRACK, RBRACK, COLON, COMMA = map(pp.Suppress, "()[]:,")

label = pp.Word(
    ppu.BasicMultilingualPlane.alphas + "_-₮.%",
    ppu.BasicMultilingualPlane.alphanums + "_-₮.% "
)


def parse_bool(tokens):
    if (len(tokens) > 1):
        raise ValueError("[parse_bool] too many tokens")
    return tokens[0] == "true"


boolean = (pp.Keyword("true") | pp.Keyword(
    "false"))("boolean").set_parse_action(parse_bool)

number = pp.common.number("number")

Ox = pp.Literal("0x")
bytes32 = pp.Combine(Ox + pp.Word(pp.hexnums, exact=64))("bytes32")
address = pp.Combine(Ox + pp.Word(pp.hexnums, exact=40))("address")
bytes16 = pp.Combine(Ox + pp.Word(pp.hexnums, exact=32))("bytes16")
bytes8 = pp.Combine(Ox + pp.Word(pp.hexnums, exact=16))("bytes8")
bytes4 = pp.Combine(Ox + pp.Word(pp.hexnums, exact=8))("bytes4")
_bytes = pp.Combine(Ox + pp.Word(pp.hexnums))("bytes")
_bytes_empty = Ox("bytes_empty")

# hex = bytes4 | bytes8 | bytes16 | address | bytes32 | _bytes
hex = (bytes32 | address | bytes16 | bytes8 | bytes4 | _bytes_empty) ^ (_bytes)

string = pp.QuotedString('"')("string")

labled_address = label("label") + COLON + LBRACK + address("address") + RBRACK

number_with_sci = number("number") + \
    pp.Optional(LBRACK + number("scientific") + RBRACK)

filtered_value = pp.Group(pp.Literal(
    "<") + label("label") + pp.Literal(">"))("filtered_value")

value = pp.Forward().set_name("value")
struct = pp.Forward().set_name("struct")

array = (
    LBRACK +
    pp.Optional(
        pp.delimited_list(pp.Group(value), delim=",")
    )("array") +
    RBRACK
)("array")

value << (
    boolean | labled_address | hex | number_with_sci |
    string | filtered_value | array | struct
)

struct_member = pp.Group(
    label("key") + COLON + pp.Group(value)("value")
)

struct_members = pp.delimited_list(struct_member).set_name(None)
struct << label("struct") + pp.Dict(
    pp.Suppress('({') +
    pp.Optional(struct_members) +
    pp.Suppress('})'),
)("members")


# struct <<= pp.Group(
#     label("struct") + pp.nested_expr(
#         '({', '})',
#         pp.dict_of(
#             label("key") + COLON,
#             pp.Group(value)("value") + pp.Optional(COMMA)))("members"))



import os
import subprocess
from collections import namedtuple

from textual import log

from crucible.context import DIR
from crucible.parser.parser import output

RunConfig = namedtuple(
    "RunConfig", [
        "profile",
        "match_contract",
        "match_test",
        "match_path",
        "verbose"
    ]
)


class ForgeRunner():
    def run(self, config: RunConfig):
        result = subprocess.run(
            ["sh", "-c", f"cd {DIR} && {self.make_command(config)}"],
            capture_output=True,  # Python >= 3.7 only
            text=True  # Python >= 3.7 only
        )
        try:
            return output.parse_string(result.stdout, parse_all=True)
        except Exception as e:
            self.save_fixture(config, result.stdout)
            raise e

    def save_fixture(self, config: RunConfig, fixture: str):
        fixture_path = f"/tmp/{config.match_contract}_{config.match_test}.txt"
        with open(fixture_path, "w") as f:
            f.write(fixture)

    def make_command(self, config: RunConfig):
        args = ["forge", "test"]
        env = os.environ.copy()
        env["FOUNDRY_PROFILE"] = config.profile
        if config.match_path:
            args.append(f"--match-path {config.match_path}")
        if config.match_contract:
            args.append(f"--match-contract {config.match_contract}")
        if config.match_test:
            args.append(f"--match-test \"^{config.match_test}\\(\\)$\"")
        if config.verbose:
            args.append("-vvvvv")
        command = " ".join(args)
        log(command)
        return command


runner = ForgeRunner()



from typing import Any, Optional
from textual import work, log
from textual.screen import Screen
from textual.widgets import Header, Footer
from textual.reactive import reactive

from crucible.data.shadow_tree import ShadowTree
from crucible.runner import RunConfig
from crucible.data.suite_tree import TestNode, SuiteTree

from .test import TestScreen
from .tree import Tree
from .run_forge_test import RunForgeTest
from .run_test_commands import RunTestCommands
from .compilation_results import CompilationResults


class SuiteScreen(Screen):
    COMMANDS = {RunTestCommands}
    BINDINGS = [
        ("r", "run_suite", "Run Suite"),
        ("ENT", "run_test", "Run Test")
    ]

    shadow_tree: Optional[ShadowTree] = None
    output: Any = reactive(None, recompose=True)

    def __init__(self, **kwargs):
        super().__init__(classes="suite-screen", **kwargs)
        self.title = "All Tests"

    async def on_mount(self) -> None:
        self.action_run_suite()

    def compose(self):
        yield Header(show_clock=True)
        if self.output and self.output.compilation.failed:
            yield CompilationResults(self.output, id="compilation")
        elif self.shadow_tree:
            yield Tree(self.shadow_tree, id="tree")
        yield Footer()

    @work
    async def action_run_suite(self) -> None:
        output = await self.app.push_screen_wait(
            RunForgeTest(RunConfig("default", None, None, None, False),
                         "full test suite")
        )

        self.output = output
        if self.output.compilation.failed:
            self.call_after_refresh(self.focus_child, "compilation")
        else:
            self.shadow_tree = SuiteTree(output)
            self.call_after_refresh(self.focus_child, "tree")

    def on_test_selected(self, evt):
        node = evt.node
        if isinstance(node, TestNode):
            if node.test.test == "setUp":
                self.app.push_screen(TestScreen(node.parent))
            else:
                self.app.push_screen(TestScreen(node))

    def on_tree_node_selected(self, evt):
        if not self.shadow_tree:
            return
        node = self.shadow_tree.get_node(evt.node.data)
        if isinstance(node, TestNode):
            log(node.test)
            if node.test.test == "setUp":
                self.app.push_screen(TestScreen(node.parent))
            else:
                self.app.push_screen(TestScreen(node))

    def focus_child(self, id) -> None:
        self.set_focus(self.get_child_by_id(id))



import tomllib
import os
# This file contains the Context class which is used to
# parse and store the context of the current execution.

# DIR = os.path.abspath("../mento/mento-core")
DIR = os.getcwd()
FOUNDRY_TOML = os.path.join(DIR, "foundry.toml")


class Context():
    def __init__(self):
        if (not os.path.exists(FOUNDRY_TOML)):
            return
        with open(FOUNDRY_TOML, "rb") as f:
            self.foundry_config = tomllib.load(f)


context = Context()



import re
from typing import TypeGuard, Iterator
from enum import Enum
from textual import log
from rich.text import Text
from .shadow_tree import ShadowNode, ShadowTree


class NodeStatus(Enum):
    FAIL = 0
    PASS = 1
    SKIP = 2


def is_suite_tree_node(node: ShadowNode) -> TypeGuard["SuiteTreeNode"]:
    return isinstance(node, SuiteTreeNode)


def has_status(status: NodeStatus):
    def __has_status__(node: SuiteTreeNode):
        return node.status == status
    return __has_status__


class SuiteTreeNode(ShadowNode):
    def __display__(self):
        if self.status == NodeStatus.PASS:
            return Text(" ", style="green") + Text(self.__title__(), style="default")
        elif self.status == NodeStatus.SKIP:
            return Text("󱔕 ", style="yellow") + Text(self.__title__(), style="light")
        elif self.status == NodeStatus.FAIL:
            return Text("❌ "+self.__title__(), style="red")
        return Text("[InvalidStatus] " + self.__title__())

    def __title__(self):
        return str(self.__id__)

    def __suite_tree_children__(self) -> Iterator["SuiteTreeNode"]:
        return filter(is_suite_tree_node, self.children)

    def __failed_children__(self) -> list["SuiteTreeNode"]:
        return list(filter(
            has_status(NodeStatus.FAIL),
            self.__suite_tree_children__()))

    def __skipped_children__(self) -> list["SuiteTreeNode"]:
        return list(filter(
            has_status(NodeStatus.SKIP),
            self.__suite_tree_children__()))

    @property
    def status(self) -> NodeStatus:
        if len(self.__failed_children__()) > 0:
            return NodeStatus.FAIL
        if len(self.__skipped_children__()) == len(self.children):
            return NodeStatus.SKIP
        return NodeStatus.PASS

    def merge_up(self, child) -> None | ShadowNode:
        pass


class SuitePathNode(SuiteTreeNode):
    def __init__(self, segment):
        super().__init__(id=segment, expanded=True)

    def merge_up(self, child):
        merged_node = None
        if isinstance(child, SuitePathNode):
            merged_node = SuitePathNode(f"{self.__id__}_{child.__id__}")
            merged_node.tree = self.tree
            merged_node.add_children_from_node(child)
        elif isinstance(child, SuiteNode):
            merged_node = SuiteNode(
                f"{self.__id__}_{child.__id__}", child.suite)
            merged_node.tree = self.tree
            merged_node.add_children_from_node(child)
        return merged_node

    def __title__(self):
        if (self.__id__.startswith("root/")):
            return str(self.__id__).replace("root/", "")
        else:
            return str(self.__id__)


class TestPathNode(SuiteTreeNode):
    def __init__(self, segment, suite):
        super().__init__(id=segment, expanded=False)
        self.suite = suite

    def merge_up(self, child):
        merged_node = None
        if isinstance(child, TestPathNode):
            merged_node = TestPathNode(
                f"{self.__id__}_{child.__id__}", self.suite)
        elif isinstance(child, TestNode):
            merged_node = TestNode(
                f"{self.__id__}_{child.__id__}", child.test, child.suite)
        if merged_node:
            merged_node.tree = self.tree
            merged_node.add_children_from_node(child)
        return merged_node


class SuiteNode(SuiteTreeNode):
    def __init__(self, segment, suite):
        super().__init__(id=segment, expanded=False)
        self.suite = suite

    def __title__(self):
        total_tests = self.suite.passed + self.suite.failed + self.suite.skipped
        return f"{self.__id__} [{self.suite.passed}/{total_tests}]"


class TestNode(SuiteTreeNode):
    def __init__(self, label, test, suite):
        super().__init__(label, expanded=False)
        self.test = test
        self.suite = suite

    @property
    def status(self):
        if self.test.result == "PASS":
            return NodeStatus.PASS
        elif self.test.result == "FAIL":
            return NodeStatus.FAIL
        elif self.test.result == "SKIP":
            return NodeStatus.SKIP
        else:
            raise ValueError(f"Unexpected result: {self.test.result}")


class SuiteTree(ShadowTree):
    def __init__(self, output):
        super().__init__()
        self.output = output
        self.set_root(SuitePathNode("root"))
        for suite in self.output.suites:
            suite_node = self.__add_suite__(suite)
            for test in suite.tests:
                self.__add_test__(test, suite, suite_node)
        self.__normalize__(self.root)
        self.__sort_children__(self.root)

    def __sort_children__(self, node):
        node.children.sort(
            key=lambda x: (x.status.value, -1 * len(x.children), x.__id__))
        for child in node.children:
            self.__sort_children__(child)

    def __normalize__(self, node, child_index=0):
        if len(node.children) == 1:
            child = node.children[0]
            merged_node = node.merge_up(child)
            if merged_node:
                if node.parent:
                    node.parent.replace_child(child_index, merged_node)
                else:
                    self.set_root(merged_node)
                self.__normalize__(merged_node, child_index)
            else:
                self.__normalize__(child, 0)
        else:
            for (i, child) in enumerate(node.children.copy()):
                self.__normalize__(child, i)

    def __add_test__(self, test, suite, suite_node):
        path = re.split(r"[:/_]", test.test)
        if path[0] == "test" or path[0] == "testFuzz" or path[0] == "invariant":
            path = path[1:]
        test_label = path.pop()
        current_node = suite_node
        for segment in path:
            found_child = False
            for child in current_node.children:
                if child.__id__ == segment and not isinstance(child, TestNode):
                    current_node = child
                    found_child = True
                    break
            if not found_child:
                current_node = current_node.add_child(
                    TestPathNode(segment, suite))
        current_node.add_child(TestNode(test_label, test, suite))

    def __add_suite__(self, suite):
        path = re.split(r"[:/_]", suite.contract)
        path = list(filter(lambda x: not x.endswith(".sol"), path))
        suite_label = path.pop()
        current_node = self.root
        for segment in path:
            found_child = False
            for child in current_node.children:
                if child.__id__ == segment:
                    current_node = child
                    found_child = True
                    break
            if not found_child:
                current_node = current_node.add_child(SuitePathNode(segment))
        return current_node.add_child(SuiteNode(suite_label, suite))



from uuid import uuid4
from textual import log
from rich.text import Text

from crucible.data.shadow_tree import ShadowNode, ShadowTree, NodeID
from crucible.data.suite_tree import TestNode, SuiteNode
from crucible.data.text_styles import default, light, keyword, ret, event, revert
from crucible.data.call_filters import call_method_eq, call_depth_lt, call_depth_eq


class TraceTreeNode(ShadowNode):
    def __init__(self, expanded=False):
        id = NodeID(str(uuid4()))
        super().__init__(id, expanded)

    def __repr__(self):
        if self.__short__():
            return str(self.__short__())
        return str(self.__display__())

    def __short__(self) -> Text | str:
        return f"<missing short representation for {self.__class__.__name__}>"


class CallNode(TraceTreeNode):
    args: list["LabeledArgumentNode"] = []

    def __init__(self, call):
        super().__init__(
            call_method_eq(call, "setUp") and
            call_depth_eq(call, 0) or
            call_depth_lt(call, 1))

        self.call = call
        self.args = []
        if (self.call.function.args):
            for i, arg in enumerate(self.call.function.args):
                arg_node = LabeledArgumentNode(arg, i.__str__())
                self.args.append(arg_node)
                self.add_child(arg_node)

    def __display__(self):
        return self.gas + " " + self.target + default("::") + self.function + self.call_type

    @property
    def gas(self):
        return light(f"[{self.call.gas}]")

    @property
    def call_type(self):
        if self.call.function.call_type:
            return light(f" [{self.call.function.call_type}]")
        else:
            return ''

    @property
    def target(self):
        if self.call.function.target.label:
            return keyword(self.call.function.target.label)
        else:
            return keyword(self.call.function.target.address)

    @property
    def function(self):
        return keyword(f"{self.method}(") + self.arguments + keyword(")")

    @property
    def arguments(self):
        if (self.args):
            output = Text()
            for i, arg in enumerate(self.args):
                output.append(arg.__short__())
                if i < len(self.args) - 1:
                    output.append(default(", "))
            return output
        elif self.call.function.raw_args:
            if len(self.call.function.raw_args[0]) > 32:
                return Text(self.call.function.raw_args[0][0:32] + "...")
            else:
                return Text(self.call.function.raw_args[0])
        else:
            return Text('')

    @property
    def method(self):
        if self.call.function.name:
            return Text(self.call.function.name)
        elif self.call.function.selector:
            return Text(self.call.function.selector)
        else:
            log(self.call)
            raise ValueError("No method name or selector found")


class ReturnNode(TraceTreeNode):
    def __init__(self, ret):
        super().__init__(False)
        self.ret = ret
        if ('returned_values' in self.ret and self.ret['returned_values']):
            for i, val in enumerate(self.ret.returned_values):
                val_node = LabeledArgumentNode(val, i.__str__())
                self.add_child(val_node)

    def __display__(self):
        return ret("← [Return]") + self.return_value

    @property
    def return_value(self):
        if ('returned_bytes' in self.ret):
            return default(" " + self.ret['returned_bytes']['count']) + light(" bytes of code")
        elif ('returned_values' in self.ret):
            output = Text(" ")
            for i, child in enumerate(self.children):
                assert isinstance(child, TraceTreeNode)
                output.append(child.__short__())
                if i < len(self.children) - 1:
                    output.append(", ")
            return output
        return ""


class RevertNode(TraceTreeNode):
    def __init__(self, revert):
        super().__init__(False)
        self.revert = revert

    def __display__(self):
        return revert("← [Revert] " + self.revert['reason'])


class StopNode(TraceTreeNode):
    def __init__(self):
        super().__init__(False)

    def __display__(self):
        return ret("← [Stop]")


class BaseValueNode():
    def __init__(self, value):
        self.value = value

    def __display__(self) -> Text | str:
        return "<Display not implemented for: " + self.__class__.__name__ + ">"

    def __short__(self) -> Text | str:
        return self.__display__()


class RawNode(BaseValueNode):
    def __display__(self):
        return "Raw: " + self.value.__str__()


class AddressNode(BaseValueNode):
    def __display__(self):
        if (self.value.label):
            output = Text()
            output.append(default(self.value.label + ": "))
            output.append(light(self.value.address))
            return output
        else:
            return str(self.value.address)

    def __short__(self):
        if (self.value.label):
            output = Text()
            output.append(default(self.value.label))
            return output
        else:
            output = Text()
            output.append(default(self.value.address[0:10] + "..."))
            return output


class NumberNode(BaseValueNode):
    def __display__(self):
        output = Text()
        output.append(default(self.value.number.__str__()))
        if (self.value.scientific):
            output.append(" ")
            output.append(light(self.value.scientific.__str__()))
        return output

    def __short__(self):
        output = Text()
        if (self.value.scientific):
            output.append(default(self.value.scientific.__str__()))
        else:
            number_str = self.value.number.__str__()
            if len(number_str) > 10:
                output.append(default(number_str[0:10] + "..."))
            else:
                output.append(default(number_str))
        return output


class StringNode(BaseValueNode):
    def __display__(self):
        return default(f"\"{self.value.string}\"")


class BytesNode(BaseValueNode):
    def __display__(self):
        output = Text()
        output.append(default(self.value[0][0:66]))
        if len(self.value[0]) > 66:
            return output.append(default("..."))
        return output

    def __short__(self):
        output = Text()
        output.append(default(self.value[0][0:10] + "..."))
        return output


class StructNode(BaseValueNode):
    def __display__(self):
        return default(self.value.struct)

    def __short__(self):
        return default(self.value.struct+"{}")


class ArrayNode(BaseValueNode):
    def __display__(self):
        return default(f"[{len(self.value.array)}]")


class BoolNode(BaseValueNode):
    def __display__(self):
        if (self.value.boolean):
            return default("true")
        else:
            return default("false")


class FilteredValueNode(BaseValueNode):
    def __display__(self):
        if (self.value.filtered_value.label):
            return light(f"<{self.value.filtered_value.label}>")
        return "<filtered-value>"


class ValueNode(TraceTreeNode):
    COLLAPSABLE = False

    def __init__(self, value):
        super().__init__(False)
        if value.string:
            self.value = StringNode(value)
        elif value.address:
            self.value = AddressNode(value)
        elif value.number != '':
            self.value = NumberNode(value)
        elif (value.bytes or value.bytes32 or value.bytes16 or value.bytes8 or value.bytes4):
            self.value = BytesNode(value)
        elif value.struct:
            self.value = StructNode(value)
            for member in value.members:
                self.children.append(
                    LabeledArgumentNode(member.value, member.key))
        elif value.get_name() == "array":
            self.value = ArrayNode(value)
            for i, item in enumerate(value.array):
                self.children.append(LabeledArgumentNode(item, i.__str__()))
        elif value.boolean != '':
            self.value = BoolNode(value)
        elif value.filtered_value:
            self.value = FilteredValueNode(value)
        else:
            self.value = RawNode(value)

    def __short__(self):
        return self.value.__short__()

    def __display__(self):
        return self.value.__display__()


class LabeledArgumentNode(ValueNode):
    __label__: str

    def __init__(self, value, label):
        super().__init__(value)
        self.__label__ = label

    def __display__(self):
        output = Text()
        output.append(light(self.__label__ + ": "))
        output.append(super().__display__())
        return output

    def __short__(self):
        return super().__short__()


class CreateNode(TraceTreeNode):
    def __init__(self, create):
        super().__init__(False)
        self.create = create

    def __display__(self):
        output = Text()
        output.append(ret("→  new "))
        output.append(ret(self.create.contract_name))
        output.append(light("@" + self.create.address))
        return self.gas + " " + output

    @property
    def gas(self):
        return light(f"[{self.create.gas}]")


class EventNode(TraceTreeNode):
    def __init__(self, event):
        super().__init__(True)
        self.event = event
        for arg in event.args:
            self.children.append(LabeledArgumentNode(arg.value, arg.key))

    def __display__(self):
        return event(f"emit {self.event.name}") + "(" + self.arguments + ")"

    @property
    def arguments(self):
        if (self.event.children):
            output = Text()
            for i, arg in enumerate(self.children):
                if (isinstance(arg, ValueNode)):
                    output.append(arg.__short__())
                    if i < len(self.children) - 1:
                        output.append(default(", "))
            return output
        else:
            return Text('')


class RawEventNode(TraceTreeNode):
    def __init__(self, event):
        super().__init__(True)
        self.event = event
        if (event.topic0):
            self.children.append(LabeledArgumentNode(event.topic0, 'topic0'))
        if (event.topic1):
            self.children.append(LabeledArgumentNode(event.topic1, 'topic1'))
        if (event.topic2):
            self.children.append(LabeledArgumentNode(event.topic2, 'topic2'))
        if (event.topic3):
            self.children.append(LabeledArgumentNode(event.topic3, 'topic3'))
        if (event.data):
            self.children.append(LabeledArgumentNode(event.data, 'data'))

    def __display__(self):
        return event("emit UnknownEvent") + "(" + self.arguments + ")"

    @property
    def arguments(self):
        if (self.event.children):
            output = Text()
            for i, arg in enumerate(self.children):
                if (isinstance(arg, ValueNode)):
                    output.append(arg.__short__())
                    if i < len(self.children) - 1:
                        output.append(default(", "))
            return output
        else:
            return Text('')


class TraceTree(ShadowTree):
    def __init__(self, output):
        super().__init__()
        self.output = output
        suite = output.suites[0]
        tests = output.suites[0].tests
        if (len(tests) > 1):
            self.set_root(SuiteNode(suite.contract, suite))
            for test in tests:
                node = self.root.add_child(
                    TestNode(test.test, test, suite))
                self.__add_test__(node, test)
        else:
            test = tests[0]
            self.set_root(TestNode(tests[0].test, tests[0], suite))
            self.__add_test__(self.root, test)

    def __add_test__(self, node, test):
        current_node = node
        for trace_group in test.trace_groups:
            for line in trace_group.traces:
                current_node = self.__add_child__(current_node, line)

    def __add_child__(self, node, line):
        if line.get_name() == "call":
            return node.add_child(CallNode(line))
        elif line.get_name() == "create":
            return node.add_child(CreateNode(line))
        elif line.get_name() == "event":
            node.add_child(EventNode(line))
            return node
        elif line.get_name() == "raw_event":
            node.add_child(RawEventNode(line))
            return node
        elif line.get_name() == "end":
            if ('return' in line):
                node.add_child(ReturnNode(line['return']))
            elif ('revert' in line):
                node.add_child(RevertNode(line['revert']))
            elif ('stop' in line):
                node.add_child(StopNode())
            else:
                log(f"Unknown `end` type: `{line}`")
            return node.parent
        else:
            log(f"Unknown line type: `{line.get_name()}`")
            return node



from typing import NewType, Callable, Optional
from rich.text import Text

NodeID = NewType("NodeID", str)
FilterKey = NewType("FilterKey", str)
Filter = Callable[["ShadowNode"], bool]


class ShadowTree():
    nodes: dict[NodeID, "ShadowNode"] = {}
    root: "ShadowNode"
    filters: dict[FilterKey, Filter] = {}
    __focused__: NodeID

    def __init__(self):
        self.filters = {}
        self.root = ShadowNode(NodeID("root"))
        self.__focused__ = self.root.id
        self.add_node(self.root)

    def set_root(self, root: "ShadowNode"):
        if (self.root):
            self.remove_node(self.root)
        self.root = root
        self.root.tree = self
        self.__focused__ = self.root.id
        self.add_node(root)

    def remove_node(self, node: "ShadowNode"):
        if (self.nodes.get(node.id)):
            self.nodes.pop(node.id)
        for child in node.children:
            self.remove_node(child)

    def add_node(self, node: "ShadowNode"):
        self.nodes[node.id] = node
        for child in node.children:
            self.add_node(child)

    def set_focused(self, id: NodeID):
        self.__focused__ = id

    @property
    def focused(self):
        focused = self.nodes[self.__focused__]
        if focused is None:
            raise ValueError("Focused node not found")
        while focused.hidden and focused.parent:
            focused = focused.parent
        return focused.id

    def get_node(self, id: NodeID):
        return self.nodes.get(id)

    def add_filter(self, key: FilterKey, filter: Filter):
        self.filters[key] = filter

    def remove_filter(self, key):
        self.filters.pop(key)

    def add_filters(self, filters: dict[FilterKey, Filter]):
        self.filters.update(filters)


class ShadowNode():
    __id__: NodeID
    expanded: bool
    children: list["ShadowNode"]

    tree: Optional[ShadowTree] = None
    parent: Optional["ShadowNode"] = None

    def __init__(self, id: NodeID, expanded: bool = False):
        self.__id__ = id
        self.expanded = expanded
        self.children = []

    @property
    def id(self):
        if self.parent:
            return self.parent.id + "::" + self.__id__
        else:
            return self.__id__

    @property
    def hidden(self):
        if self.tree:
            for filter in self.tree.filters.values():
                if filter(self):
                    return True
        return False

    def toggle_expanded(self, expanded):
        self.expanded = expanded

    def add_children_from_node(self, node: "ShadowNode"):
        for child in node.children:
            if self.tree:
                self.tree.remove_node(child)
            self.add_child(child)
        node.children = []

    def add_child(self, child: "ShadowNode"):
        self.children.append(child)
        child.parent = self
        if (self.tree):
            child.tree = self.tree
            self.tree.add_node(child)
        return child

    def replace_child(self, index, child: "ShadowNode"):
        old_child = self.children[index]
        self.children[index] = child
        child.parent = self
        child.tree = self.tree
        if self.tree:
            self.tree.remove_node(old_child)
            self.tree.add_node(child)

    @property
    def label(self):
        return self.__display__()

    def __display__(self) -> str | Text:
        return "__display__ not implemented for:  " + str(self)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.id}>"



from textual.app import App
from screens.suite import SuiteScreen


class Crucible(App):
    CSS_PATH = "crucible.tcss"
    BINDINGS = [
        ("q", "quit", "Quit"),
    ]
    MODES = {
        "app": SuiteScreen
    }

    def on_mount(self) -> None:
        self.switch_mode("app")

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark


def crucible():
    Crucible().run()


if __name__ == "__main__":
    crucible()
